package com.github.jfcloud.excel.editor.docdeal.controller;


import cn.hutool.core.io.IoUtil;
import cn.hutool.core.io.file.FileNameUtil;
import cn.hutool.core.lang.Pair;
import cn.hutool.core.net.URLEncodeUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.http.HttpResponse;
import cn.hutool.http.HttpUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.amazonaws.services.s3.model.S3Object;
import com.github.jfcloud.excel.editor.docdeal.bean.*;
import com.github.jfcloud.excel.editor.docdeal.constant.CommandEnum;
import com.github.jfcloud.excel.editor.docdeal.constant.DocumentConstants;
import com.github.jfcloud.excel.editor.docdeal.constant.DocumentStatus;
import com.github.jfcloud.excel.editor.docdeal.oss.OssProperties;
import com.github.jfcloud.excel.editor.docdeal.oss.service.OssTemplate;
import com.github.jfcloud.excel.editor.docdeal.service.FileUploadService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

@Slf4j
@Controller
public class FileController {
    @Autowired
    RestTemplate restTemplate;
    @Autowired
    OssTemplate ossTemplate;
    @Autowired
    OssProperties ossProperties;
    @Value("${files.docservice.url.site}")
    private String officeUrl;
    @Value("${files.docservice.url.command}")
    private String officeCommand;
    @Value("${document.server.url}")
    private String serverUrl;

    @Autowired
    private FileUploadService uploadService;

    @ResponseBody
    @PostMapping(value = "upload")
    public String upload(@RequestPart("file") MultipartFile file, @RequestParam(required = false) String bucket) throws Exception {
        if (file.isEmpty()) {
            throw new Exception("上传文件不能为空");
        }

        String fileName = file.getOriginalFilename();
        log.info("上传文件 {}", fileName);

        if (StrUtil.isBlank(bucket)) {
            bucket = ossProperties.getBucketName();
        }
        ossTemplate.putObject(bucket, fileName, file.getInputStream());
        log.info("上传文档已保存，bucket={} name={}", bucket, fileName);

        FileUpload upload = new FileUpload();
        upload.setUpload_date(new Date());
        upload.setFile_type(FileNameUtil.getSuffix(fileName));
        upload.setFile_path(String.format("/admin/sys-file/%s/%s", bucket, fileName));
        upload.setFile_name(fileName);
        upload.setFile_size(file.getSize());
        uploadService.save(upload);

        return "上传成功";
    }

    /**
     * 查询所有上传文档信息接口
     *
     * @return
     */
    @GetMapping("/filelist")
    @ResponseBody
    public List<FileUpload> listFile() {
        return uploadService.list();
    }

    /**
     * 下载文档接口
     *
     * @param name
     * @param response
     */
    @GetMapping("/download")
    public void download(@RequestParam String name,
                         @RequestParam(required = false) String bucket,
                         HttpServletResponse response) {
        if (StrUtil.isBlank(bucket)) {
            bucket = ossProperties.getBucketName();
        }
        log.info("==> 下载文档: name={}, bucket={}", name, bucket);

        long size = ossTemplate.getSize(bucket, name);

        try {
            if (size < 0) {
                log.error("文档不存在 name={}, bucket={}", name, bucket);
                response.getWriter().println("Sorry, we cannot find the file, pls check the file name and try again.");
                return;
            }

            response.setContentType("application/octet-stream");
            response.setHeader("Content-Disposition", "attachment;filename*=utf8''" + URLEncodeUtil.encode(name));
            response.setHeader("Content-Length", String.valueOf(size));

            S3Object s3Object = ossTemplate.getObject(bucket, name);
            IoUtil.copy(s3Object.getObjectContent(), response.getOutputStream());
        } catch (IOException e) {
            log.error("文档下载失败 name={}, error={}", name, e.getMessage());
        }
    }

    @GetMapping("/review")
    public String reviewDocFile(@RequestParam(required = false) String bucket,
                                @RequestParam(required = false) String title,
                                @RequestParam String name,
                                @RequestParam String userId,
                                @RequestParam String userName,
                                Model model) {
        log.info("==> 预览文档 bucket: {}, name: {}", bucket, name);

        Pair<Document, DocumentEditParam> pair = docPair(bucket, title, name, userId, userName);
        Document document = pair.getKey();
        model.addAttribute("document", document);
        model.addAttribute("documentEditParam", pair.getValue());

        if (document.getFileType().equals("xls") || document.getFileType().equals("xlsx")) {
            return "/viewerExcel";
        }
        return "/viewer";
    }

    private Pair<Document, DocumentEditParam> docPair(String bucket, String title, String name,
                                                      String userId, String userName) {

        title = StrUtil.isBlank(title) ? name : title;
        bucket = StrUtil.isBlank(bucket) ? ossProperties.getBucketName() : bucket;

        String downloadUrl = String.format(DocumentConstants.OFFICE_API_DOC_FILE, serverUrl, name);
        downloadUrl += "&bucket=" + bucket;

        Document document = Document.builder()
                .key(FileNameUtil.getPrefix(name))
                .title(title)
                .fileType(FileNameUtil.getSuffix(name))
                .url(downloadUrl)
                .permissions(new JSONObject()
                        .fluentPut("chat", true)
                        .fluentPut("comment", true)
                        .fluentPut("copy", true)
                        .fluentPut("download", true)
                        .fluentPut("edit", false)
                        .fluentPut("fillForms", false)
                        .fluentPut("modifyContentControl", true)
                        .fluentPut("modifyFilter", true)
                        .fluentPut("print", true)
                        .fluentPut("review", false)
                        .fluentPut("reviewGroups", null)
                        .fluentPut("commentGroups", null)
                        .fluentPut("userInfoGroups", null))
                .build();
        log.info("document: {}", JSON.toJSONString(document));

        DocumentEditParam param = DocumentEditParam.builder()
                .callbackUrl(String.format(DocumentConstants.OFFICE_API_CALLBACK_BUCKET, serverUrl, name, bucket))
                .user(new DocumentEditParam.UserBean(userId, userName))
                .build();
        log.info("param: {}", JSON.toJSONString(param));

        return Pair.of(document, param);
    }

    @GetMapping("/reviewExcel")
    public String reviewExcel(@RequestParam(required = false) String bucket,
                              @RequestParam(required = false) String title,
                              @RequestParam String name,
                              String userId,
                              String userName,
                              Model model) {
        return reviewDocFile(bucket, title, name, userName, userId, model);
    }

    @GetMapping("/edit")
    public String editDocFile(@RequestParam(required = false) String bucket,
                              @RequestParam(required = false) String title,
                              @RequestParam String name,
                              String userName,
                              String userId,
                              Model model) {
        log.info("==> 编辑文档 bucket: {}, name: {}", bucket, name);

        Pair<Document, DocumentEditParam> pair = docPair(bucket, title, name, userId, userName);
        Document document = pair.getKey();
        document.getPermissions().clear();
        model.addAttribute("document", document);
        model.addAttribute("documentEditParam", pair.getValue());
        return "/editor";
    }


    /**
     * Only Office 回调接口
     * <a href="https://api.onlyoffice.com/zh-CN/docs/docs-api/usage-api/callback-handler/">回调参数</a>
     */
    @RequestMapping("/callback")
    @ResponseBody
    public DocumentEditCallbackResponse callback(@RequestBody DocumentEditCallback param, HttpServletRequest request) {
        String name = request.getParameter("name");
        String bucket = request.getParameter("bucket");
        log.info("==> onlyoffice 回调 name={}, bucket={}, body={}", name, bucket, JSON.toJSONString(param));

        String userIds = param.getActions()
                .stream()
                .map(DocumentEditCallback.ActionsBean::getUserid)
                .collect(Collectors.joining(", "));

        try {
            switch (param.getStatus()) {
                case 0:
                    log.info("找不到具有对应key的文档 user={}", userIds);
                    break;
                case 1:
                    log.info("文档正在被编辑 user={}", userIds);
                    break;
                case 2:
                case 6:
                    if (param.getStatus() == 2) {
                        log.info("文档已准备好保存 user={}", userIds);
                    } else {
                        log.info("正在进行强制保存请求 user={}", userIds);
                    }

                    //更新oss文件
                    byte[] bytes = HttpUtil.downloadBytes(param.getUrl());
                    ossTemplate.putObject(bucket, name, new ByteArrayInputStream(bytes));
                    log.info("文档已保存，bucket={} name={}", bucket, name);
                    break;
                case 3:
                    log.info("文档保存时发生错误 user={}", userIds);
                    break;
                case 4:
                    log.info("文档关闭时没有更改 user={}", userIds);
                    break;
                case 7:
                    log.info("强制保存文档时发生错误 user={}", userIds);
                    break;
                default:
                    log.info("未知状态 user={}", userIds);
            }
        } catch (Exception ex) {
            log.error("文档保存时发生错误 user={} msg={}", userIds, ex.getMessage());
            return DocumentEditCallbackResponse.fail();
        }
        return DocumentEditCallbackResponse.success();
    }

    /**
     * 获取文档编辑状态
     *
     * @param name
     * @return
     */
    @GetMapping("/editStatus")
    @ResponseBody
    public JSONObject getDoucmentEditStatus(@RequestParam String name, @RequestParam(required = false) String command) {
        log.info("查询文档编辑状态 name={}", name);

        command = StrUtil.isBlank(command) ? CommandEnum.FORCE_SAVE.getCommand() : command;
        JSONObject params = new JSONObject()
                .fluentPut("c", command)
                .fluentPut("key", name);

        try (HttpResponse response = HttpUtil.createPost( officeUrl + officeCommand)
                .body(params.toJSONString())
                .header("Content-Type", "application/json")
                .execute()) {

            JSONObject responseObj = JSON.parseObject(response.body());
            if (CommandEnum.INFO.getCommand().equals(command)) {
                DocumentStatus documentStatus = DocumentStatus.getDocumentStatus(responseObj.getInteger("error"));
                if (documentStatus != null) {
                    responseObj.put("msg", documentStatus.getMsg());
                }
            }

            log.info("文档（{}）编辑状态 {}", name, responseObj);
            return responseObj;
        }
    }

}
